/* 
 *  rocks/reconnect.c
 *
 *  Reconnection procedure.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */
/* 
   Overview of reconnection procedure:

   We open a passive socket on the address that the peer last knew
   about (rs->sa_locl), and we try to connect to the peer address that
   we last knew about (rs->sa_peer).  If at least one of us can bind
   to our last known address, eventually we'll reconnect.  If we can
   both reconnect, we'll choose one of the connections and drop the
   other.

   rs_reconnect does not return until it has established a connection.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>

#include "rs.h"
#include "rstypes.h"
#include "flight.h"
#include "log.h"

/* For timing */
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
static const int DOTIMING = 1;
static struct timeval tv_start;
static struct timeval tv_end;

#define ACTIVE   0x01
#define PASSIVE  0x02
#define TIE      0x00
#define BUG      0x11
#define SET(bit, est)       ((est) |= (bit))
#define CLR(bit, est)       ((est) &= ~(bit))
#define ISSET(bit, est)     ((est) & (bit))
#define ISSETUP(s) ((s) >= 0)
#define ESTEXCHANGED(rec)  ((rec)->sent_est && (rec)->recv_est)

/* list of reliable sockets currently being reconnected */
static rs_t rec_list = NULL;

static int reconnecting = 0;

static void do_reconnect(int sd);

static void
recstate_init(recstate_t *rec)
{
	bzero(rec, sizeof(recstate_t));
	rec->sda = -1;
	rec->sdp = -1;
	rec->sdn = -1;
}

/* If there is a failure in this operation (which there shouldn't be
   unless the computer is badly configured or memory is tight), then
   silently give up.  The socket will stay open, which may delay or
   preclude reconnection, but it's safe. */
static void
socket_to_devnull(int sd)
{
	int fd;
	rs_mode_native();
	fd = open("/dev/null", O_RDONLY);
	if (0 > fd)
		goto out; /* weird */
	if (sd == fd)
		goto out; /* weird */
	close(sd);
	if (0 > dup2(fd, sd))
		goto out; /* weird */
	close(fd);
out:
	rs_mode_pop();
}

void
rs_begin_reconnect(rs_t rs)
{
	sigset_t hbsigs;
	rs_t q;

	assert(rs);
	rs_stop_heartbeat(&hbsigs);
	rs->state = RS_SUSPENDED;
	if (!rs_opt_nohb)
		rs_hb_cancel(rs->hb);
	socket_to_devnull(rs->sd);

	for (q = rec_list; q; q = q->rec_next)
		/* FIXME: Why is it legitimate for this to happen? */
		if (q == rs) {
			fprintf(stderr, "RS reconnection double entry\n");
			return;
		}
	rs->rec_next = rec_list;
	rs->rec_prev = NULL;
	if (rec_list)
		rec_list->rec_prev = rs;
	rec_list = rs;

	recstate_init(&rs->rec);
	rs_resume_heartbeat(&hbsigs);
	/* FIXME: For now, but later make this an option */
	do_reconnect(rs->sd);
}

static void
end_reconnect(rs_t rs)
{
	int rv;
	sigset_t hbsigs;
	socklen_t len;

	assert(rs);

	/* Authenticate */
#ifndef NO_AUTH
	if (!rs_opt_noauth) {
		rs_log("Reconnect: Authenticating");
		rs_mode_native();
		rv = rs_authenticate(rs->key, rs->rec.sdc);
		rs_mode_pop();
		if (0 > rv) {
			rs_log("Authentication of incoming connection failed");
			close(rs->rec.sdc);
			rs->rec.sdc = -1;
			goto failed;
		}
	}
#endif
	rs_stop_heartbeat(&hbsigs);
	assert(rs->state == RS_SUSPENDED);

	/* Dup chosen socket to reliable socket descriptor */
	if (rs->rec.sdc != rs->sd) {
		close(rs->sd);
		rv = dup2(rs->rec.sdc, rs->sd);
		assert(rv == rs->sd);
	}

	/* Remove from REC list. */
	if (rs->rec_prev)
		rs->rec_prev->rec_next = rs->rec_next;
	if (rs->rec_next)
		rs->rec_next->rec_prev = rs->rec_prev;
	if (rs == rec_list)
		rec_list = rs->rec_next;

	/* Heartbeat must be established and beating before in-flight
	   recovery so that full fault detection is available during
	   in flight recovery. */
	if (!rs_opt_nohb)
		if (0 > rs_hb_establish(rs->sd, rs->hb, rs->role))
			goto failed;
	rs->state = RS_ESTABLISHED;
	rs_resume_heartbeat(&hbsigs);

	/* FIXME: Mechanisms should either be in charge of changing the
	   mode, or not, but not both (cf. heartbeat). */
	if (!rs_opt_noflight) {
		rs_mode_native();
		rv = rs_inflight_recover(rs->sd, rs->ring,
					 rs->rcvseq, rs->sndseq,
					 &rs->maxsnd, &rs->maxrcv);
		rs_mode_pop();
		if (0 > rv)
			goto failed;
	}

	/* Update addresses */
	len = sizeof(rs->sa_locl);
	if (0 > getsockname(rs->sd, (struct sockaddr *) &rs->sa_locl, &len))
		goto failed;
	len = sizeof(rs->sa_locl);
	if (0 > getpeername(rs->sd, (struct sockaddr *) &rs->sa_peer, &len))
		goto failed;

	return;
failed:
	rs_log("reconnect failed -> begin reconnect");
	rs_reconnect(rs);
}

static void
finalize(rs_t rs, int which)
{
	assert(PASSIVE == which || ACTIVE == which);

	rs->rec.sdc = (PASSIVE == which) ? rs->rec.sdn : rs->rec.sda;
	close((PASSIVE == which) ? rs->rec.sda : rs->rec.sdn);
	close(rs->rec.sdp);
	end_reconnect(rs);
}

static void
establish(rs_t rs, int which)
{
	char byte;
	int *fd;
	int rv;

	assert(PASSIVE == which || ACTIVE == which);
	fd = (PASSIVE == which) ? &rs->rec.sdn : &rs->rec.sda;

	SET(which, rs->rec.est);

	if (0 > rs_nonblock(*fd, 0))
		assert(0);
	if (0 > rs_reset_on_close(*fd, 1))
		assert(0);

	if (RS_ROLE_CLIENT == rs->role)
		/* Wait (via select) for a byte from the peer. */
		return;

	/* Send byte to peer if we haven't already. */
	if (rs->rec.sent)
		return;
	rv = write(*fd, &byte, 1);
	if (rv != 1) {
		/* Uh oh. */
		rs_log("Reconnect: Warning: Can't send byte to client: %s",
		       strerror(errno));
		close(*fd);
		*fd = -1;
		CLR(which, rs->rec.est);
		return;
	}
	SET(which, rs->rec.sent);
	finalize(rs, which);
}

/* Setup a passive socket listening on RS->SA_LOCL.  Return the socket
   descriptor if successful, or -1 on error. */
static int
setup_passive(rs_t rs)
{
	int sd;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd)
		return -1;

	/* Non blocking avoids the timing problem described in
	   Stevens Network Programming V1, sec. 15.6 */
	if (0 > rs_nonblock(sd, 1))
		assert(0);
	if (0 > rs_reuseaddr(sd))
		assert(0);
	if (0 > bind(sd, (struct sockaddr *) &rs->sa_locl,
		     sizeof(rs->sa_locl))) {
		close(sd);
		return -1;
	} 
	if (0 > listen(sd, 1)) {
		close(sd);
		return -1;
	}
	return sd;
}

static int
setup_active(rs_t rs)
{
	int sd, rv;
	struct sockaddr_in locl;
	int len = sizeof(locl);

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd)
		return -1;
	if (0 > rs_nonblock(sd, 1))
		assert(0);
	if (0 > rs_reuseaddr(sd))
		assert(0);
	rv = connect(sd, (struct sockaddr *)&rs->sa_peer, sizeof(rs->sa_peer));
	if (0 > getsockname(sd, (struct sockaddr *)&locl, &len)) {
		close(sd);
		return -1;
	}
	if (locl.sin_port == rs->sa_peer.sin_port
	    && !memcmp(&locl.sin_addr, &rs->sa_peer.sin_addr,
		       sizeof(locl.sin_addr))) {
		/* This socket is connecting to itself. */
		close(sd);
		return -1;
	}
	if (0 == rv) {
		establish(rs, ACTIVE);
	} else if (errno != EINPROGRESS) {
		close(sd);
		return -1;
	}
	return sd;
}

static int
preselect(rs_t rs, fd_set *rset, fd_set *wset)
{
	int maxfd = -1;
	recstate_t *rec = &rs->rec;

	if (!ISSETUP(rec->sdp))
		rec->sdp = setup_passive(rs);
	if (!ISSETUP(rec->sda))
		rec->sda = setup_active(rs);
	
	/* Check either accepted incoming connection (if there is one),
	   or the passive listening socket (if there is one). */
	if (ISSETUP(rec->sdn)) {
		FD_SET(rec->sdn, rset);
		if (rec->sdn > maxfd)
			maxfd = rec->sdn;
	} else if (ISSETUP(rec->sdp)) {
		FD_SET(rec->sdp, rset);
		if (rec->sdp > maxfd)
			maxfd = rec->sdp;
	}

	/* Check the active connection (if there is one). */
	if (ISSETUP(rec->sda)) {
		FD_SET(rec->sda, rset);
		if (!ISSET(ACTIVE, rec->est))
                        /* Check write for non-blocking connect failures */
			FD_SET(rec->sda, wset); 
		if (rec->sda > maxfd)
			maxfd = rec->sda;
	}
	return maxfd;
}

static void
passive_establish(rs_t rs)
{
	struct sockaddr_in addr;
	int len = sizeof(addr);

	/* Get connection. */
	rs->rec.sdn = accept(rs->rec.sdp, (struct sockaddr *)&addr, &len);
	if (0 > rs->rec.sdn)
		return;
	if (0 > rs_reuseaddr(rs->rec.sdn))
		assert(0);
	establish(rs, PASSIVE);
}

static void
active_establish(rs_t rs)
{
	struct sockaddr_in addr;
	int len = sizeof(addr);

	/* Verify that a connection is established. */
	if (0 > getpeername(rs->rec.sda, (struct sockaddr *)&addr, &len)) {
		close(rs->rec.sda);
		rs->rec.sda = -1;
		return;
	}
	establish(rs, ACTIVE);
}

static void
receive_byte(rs_t rs, int which)
{
	char byte;
	int *fd;
	int rv;

	assert(PASSIVE == which || ACTIVE == which);
	fd = (PASSIVE == which) ? &rs->rec.sdn : &rs->rec.sda;

	rv = read(*fd, &byte, 1);
	if (0 >= rv) {
		close(*fd);
		*fd = -1;
		CLR(which, rs->rec.est);
		return;
	}
	finalize(rs, which);
}

static void
postselect(rs_t rs, fd_set *rset, fd_set *wset, int timedout)
{
	if (timedout) {
		/* FIXME: Here might be where we expire reconnection
		   attempts. */
		++rs->rec.timeouts;
		if (rs->rec.timeouts > rs_opt_rec_max_timeouts) {
			close(rs->rec.sda);
			rs->rec.sda = -1;
			CLR(ACTIVE, rs->rec.est);
			close(rs->rec.sdp);
			rs->rec.sdp = -1;
			CLR(PASSIVE, rs->rec.est);
			rs->rec.timeouts = 0;
		}
		return;
	}

	rs->rec.timeouts = 0;
	/* Check active connection establishment */
	if (ISSETUP(rs->rec.sda)
	    && !ISSET(ACTIVE, rs->rec.est)
	    && (FD_ISSET(rs->rec.sda, rset)
		|| FD_ISSET(rs->rec.sda, wset)))
		active_establish(rs);

	/* Check passive connection establishment */
	if (ISSETUP(rs->rec.sdp)
	    && !ISSET(PASSIVE, rs->rec.est)
	    && FD_ISSET(rs->rec.sdp, rset))
		passive_establish(rs);

	if (RS_ROLE_SERVER == rs->role)
		return;

	/* Check for a byte waiting on established connections */
	if (ISSETUP(rs->rec.sda)
	    && ISSET(ACTIVE, rs->rec.est)
	    && FD_ISSET(rs->rec.sda, rset))
		receive_byte(rs, ACTIVE);

	if (ISSETUP(rs->rec.sdn)
	    && ISSET(PASSIVE, rs->rec.est)
	    && FD_ISSET(rs->rec.sdn, rset))
		receive_byte(rs, PASSIVE);
}

/* This is the main routine for the reconnection procedure */
static void
do_reconnect(int blocked_sd)
{
	int rv, maxfd;
	fd_set rset, wset;
	struct timeval tv;
	int timedout;
	rs_t rs;

	if (DOTIMING)
		gettimeofday(&tv_start, NULL);

	/* FIXME: Block signals? */
	/* No reentry */
	if (reconnecting)
		return;
	reconnecting = 1;
	tv.tv_sec = tv.tv_usec = 0;
	rs_log("Connection failure.  Reconnecting.");
	rs_mode_push(RS_MODE_NATIVE);
	while ((rs_lookup(blocked_sd))->state == RS_SUSPENDED) {
		/* Setup for call to select */
		FD_ZERO(&rset);
		FD_ZERO(&wset);
		maxfd = -1;

		for (rs = rec_list; rs; rs = rs->rec_next) {
			rv = preselect(rs, &rset, &wset);
			if (rv > maxfd)
				maxfd = rv;
		}

		/* Sleep off remaining time from select call to avoid
		   becoming CPU bound. */
		if (tv.tv_usec > 0)
			select(0, NULL, NULL, NULL, &tv);
	reselect:
		tv.tv_sec = 0;
		tv.tv_usec = 500000; /* 0.5 sec */
		rv = select(maxfd+1, &rset, &wset, NULL, &tv);
		if (0 > rv)
			goto reselect;
		timedout = (rv == 0);
		
		/* Handle select results */
		rs = rec_list;
		for (rs = rec_list; rs; rs = rs->rec_next)
			postselect(rs, &rset, &wset, timedout);
	}
	rs_mode_pop();
	reconnecting = 0;
	rs_log("Connection resumed.");
	if (DOTIMING) {
		gettimeofday(&tv_end, NULL);
		rs_log("Time of day at reconnection: %ld sec %ld usec",
			tv_end.tv_sec, tv_end.tv_usec);
	}
}

void
rs_reconnect(rs_t rs)
{
	rs_begin_reconnect(rs);
}
